<?php

declare(strict_types=1);

namespace Dedi\SyliusMigrationPlugin\Service\Sql;

use Dedi\SyliusMigrationPlugin\Config\MigrationConfig;
use Dedi\SyliusMigrationPlugin\Definition\ColumnMappingDefinition;
use Dedi\SyliusMigrationPlugin\Definition\TableMappingDefinition;
use Dedi\SyliusMigrationPlugin\Registry\ErrorHandlerRegistry;
use Dedi\SyliusMigrationPlugin\Registry\TransformerRegistry;
use Dedi\SyliusMigrationPlugin\Repository\MappingRepository;
use Dedi\SyliusMigrationPlugin\Service\Logger\MigrationLoggerInterface;
use Dedi\SyliusMigrationPlugin\Service\Specification\IsIdMappedAndNotGenerated;
use Doctrine\DBAL\Connection;
use Doctrine\DBAL\Exception;
use LogicException;
use Psr\Log\LogLevel;
use Symfony\Contracts\Cache\CacheInterface;

class Writer extends AbstractSqlProcessor implements WriterInterface
{
    private readonly ErrorHandlerRegistry $errorHandlerRegistry;

    private readonly MappingRepository $mappingRepository;

    private IsIdMappedAndNotGenerated $isIdMappedAndNotGenerated;

    public function __construct(
        MigrationConfig $config,
        Connection $connection,
        TransformerRegistry $transformerRegistry,
        MigrationLoggerInterface $logger,
        ErrorHandlerRegistry $errorHandlerRegistry,
        MappingRepository $mappingRepository,
        CacheInterface $cache,
        IsIdMappedAndNotGenerated $idResolved,
    ) {
        parent::__construct($config, $connection, $transformerRegistry, $cache, $logger);
        $this->errorHandlerRegistry = $errorHandlerRegistry;
        $this->mappingRepository = $mappingRepository;
        $this->isIdMappedAndNotGenerated = $idResolved;
    }

    public function write(TableMappingDefinition $tableMappingDefinition, $data): void
    {
        $destQueryBuilder = $this->connection->createQueryBuilder();

        $existingRows = $this->handleSearch($tableMappingDefinition, $data);

        $isUpdate = null !== $existingRows;

        if ($isUpdate) {
            $destQueryBuilder->update($tableMappingDefinition->getTargetTableName());

            foreach ($existingRows as $key => $existingRow) {
                $destQueryBuilder
                    ->andWhere("{$existingRow['dest_column_name']} = :{$existingRow['dest_column_name']}{$key}")
                    ->setParameter($existingRow['dest_column_name'] . $key, $existingRow['dest_id'])
                ;
            }
        } else {
            $destQueryBuilder
                ->insert($tableMappingDefinition->getTargetTableName())
            ;
        }

        $resolvedData = $data;

        $hasColumns = !empty($tableMappingDefinition->getColumnMappingDefinitions());
        foreach ($tableMappingDefinition->getColumnMappingDefinitions() as $columnMappingDefinition) {
            $function = $isUpdate ? 'set' : 'setValue';

            call_user_func(
                [$destQueryBuilder, $function],
                $columnMappingDefinition->getTargetColumName(),
                sprintf(':%s', $columnMappingDefinition->getTargetColumName()),
            );

            $resolvedValue = $this->resolveMappingValue($columnMappingDefinition, $data);

            $destQueryBuilder
                ->setParameter($columnMappingDefinition->getTargetColumName(), $resolvedValue)
            ;

            $resolvedData[$columnMappingDefinition->getTargetColumName()] = $resolvedValue;
        }

        $this->logger->debug($tableMappingDefinition, $destQueryBuilder->getSQL(), $data);

        if (!$this->config->isCommit()) {
            return;
        }

        try {
            if (!$isUpdate || $hasColumns) {
                $destQueryBuilder->execute();
            }
            $id = $this->connection->lastInsertId();
        } catch (\Exception $exception) {
            $doLog = count($tableMappingDefinition->getErrorHandlers()) === 0;

            foreach ($tableMappingDefinition->getErrorHandlers() as $errorHandlerType => $configuration) {
                $errorHandler = $this->errorHandlerRegistry->getErrorHandler($errorHandlerType);
                if ($errorHandler->can($exception)) {
                    $errorHandler->setConfiguration($configuration['configuration']);
                    $errorHandler->execute($tableMappingDefinition, $resolvedData);

                    if (true === $errorHandler->getErrorBubbling()) {
                        $doLog = true;
                    }
                }
            }

            if ($doLog) {
                throw $exception;
            }

            return;
        }

        if (null === $existingRows) {
            if ($this->isIdMappedAndNotGenerated->isSatisfiedBy($tableMappingDefinition, $id)) {
                // Probably a case were we don't rely on the autoincrement feature to determine the id
                $resolvedData['dest_id'] = $resolvedData[$tableMappingDefinition->getIdentifierMappingDefinitions()[0]->getDestColumnName()];
            } else {
                $resolvedData['dest_id'] = $id;
            }
            $this->mappingRepository->save($tableMappingDefinition, $resolvedData);
        }
    }

    /**
     * @throws \Doctrine\DBAL\Driver\Exception
     * @throws Exception
     */
    private function resolveMappingValue(ColumnMappingDefinition $columnMappingDefinition, array $data)
    {
        $value = $data[$columnMappingDefinition->getSourceColumnName()] ?? null;

        if (null !== $columnMappingDefinition->getValue()) {
            return $columnMappingDefinition->getValue();
        }

        if (null !== $columnMappingDefinition->getAlias()) {
            $value = $data[$columnMappingDefinition->getAlias()] ?? null;
        }

        if (null !== $columnMappingDefinition->getIdentifierMapping()) {
            $value = $this->getIdentifierMappingValue($columnMappingDefinition->getIdentifierMapping(), $columnMappingDefinition->getSourceColumnName(), $data);
        }

        if (null !== $columnMappingDefinition->getSubQuerySql()) {
            $value = $this->connection->executeQuery($columnMappingDefinition->getSubQuerySql(), $data)->fetchOne();

            if (false === $value) {
                $value = null;
            }
        }

        if (null !== $columnMappingDefinition->getTransformer()) {
            $transformer = $this->transformerRegistry->getTransformer($columnMappingDefinition->getTransformer());
            $value = $transformer->execute($value);
        }

        return $value;
    }

    /**
     * @throws Exception
     */
    private function handleSearch(TableMappingDefinition $tableMappingDefinition, array $data): ?array
    {
        if (!$tableMappingDefinition->hasSearchMapping()) {
            return null;
        }

        $results = [];

        foreach ($tableMappingDefinition->getSearchMapping() as $searchMappingDefinition) {
            $value = $this->prepareDataForSearch($searchMappingDefinition->getType(), $data[$searchMappingDefinition->getValueColum()]);

            $result = $this->mappingRepository->getOne(
                $searchMappingDefinition->getSourceTable(),
                $searchMappingDefinition->getValueColum(),
                (string) $value,
                $tableMappingDefinition->getTargetTableName(),
            );

            if (false === $result) {
                return null;
            }

            $checkQueryBuilder = $this->connection->createQueryBuilder();
            $checkQueryBuilder->select('COUNT(*)')->from($tableMappingDefinition->getTargetTableName());
            $checkQueryBuilder
                ->andWhere("{$result['dest_column_name']} = :{$result['dest_column_name']}")
                ->setParameter($result['dest_column_name'], $result['dest_id'])
            ;

            $checkedLines = $checkQueryBuilder->execute();
            if (false === $checkedLines->fetchOne()) {
                try {
                    $this->mappingRepository->delete($result['source_table_name'], $result['source_column_name'], $result['dest_table_name'], $result['dest_column_name'], $result['source_id'], $result['dest_id']);
                    $this->logger->info($tableMappingDefinition, sprintf('Deleted mapping %s', json_encode($result)));
                } catch (\Exception $exception) {
                    $this->logger->warning($tableMappingDefinition, sprintf('Error %s while deleting mapping %s', $exception->getMessage(), json_encode($result)));
                }

                continue;
            }

            $results[] = $result;
        }

        return $results;
    }

    private function prepareDataForSearch(string $type, $value)
    {
        switch ($type) {
            case 'comparison':
                return $value;
            case 'like':
                return sprintf('%%%s%%', $value);
            default:
                throw new LogicException(sprintf('Unsupported type %s', $type));
        }
    }

    private function getIdentifierMappingValue(array $identifierMapping, string $sourceColumnName, array $data)
    {
        $result = $this->mappingRepository->getOne(
            $identifierMapping['source_table'],
            $identifierMapping['source_column'],
            (string) $data[$sourceColumnName],
            $identifierMapping['dest_table'],
        );

        if (false === $result) {
            return null;
        }

        return $result['dest_id'];
    }

    public function afterMigrate(TableMappingDefinition $tableMappingDefinition): void
    {
        $afterMigrateDefinition = $tableMappingDefinition->getAfterMigrateDefinition();

        if ($afterMigrateDefinition === null) {
            return;
        }

        foreach ($afterMigrateDefinition->getSql() as $sql) {
            $this->logger->log(LogLevel::DEBUG, $sql);
            if ($this->config->isCommit()) {
                $this->connection->executeQuery($sql);
            }
        }
    }
}
